from .prosit import PrositRetentionTimePredictor
from .base import RetentionTimePredictor
from .deepLC import DeepLCRetentionTimePredictor

__all__ = [
    RetentionTimePredictor,
    PrositRetentionTimePredictor,
    DeepLCRetentionTimePredictor
]


from os.path import join
from matplotlib import pyplot as plt
from dlomix.reports.Report import Report
import numpy as np
import pandas as pd

class RetentionTimeReport(Report):
    def __init__(self, output_path, history, figures_ext='pdf'):
        super(RetentionTimeReport, self).__init__(output_path, history, figures_ext)

    def generate_report(self, targets, predictions, **kwargs):
        r2 = self.calculate_r2(targets, predictions)
        self.plot_all_metrics()
        self.plot_highlight_data_portion(targets, predictions)

        # TODO: find best way to export a document or a txt file with all the results or combine with figures
        # in a pdf or something similar

    '''
    Calculate R-squared using sklearn given true targets and predictions 
    '''

    def calculate_r2(self, targets, predictions):
        from sklearn.metrics import r2_score

        r2 = r2_score(np.ravel(targets), np.ravel(predictions))
        return r2

    '''
    
    Plot histogram of residuals
    
    '''

    def plot_residuals(self, targets, predictions, xrange=(-10, 10)):
        error = np.ravel(targets) - np.ravel(predictions)

        bins = np.linspace(xrange[0], xrange[1], 200)

        plt.hist(error, bins, alpha=0.5, color="orange")
        plt.title("Historgram of Residuals")
        plt.xlabel("Residual value")
        plt.ylabel("Count")
        plt.show()

    '''
    Plot results and highlight a portion of the data (e.g. 95%)  given true targets and predictions 
    '''

    def plot_highlight_data_portion(self, targets, predictions, portion=0.95):
        # 95% percent of the data-points highlighted

        df = pd.DataFrame({'preds': np.ravel(predictions), 'y': np.ravel(targets)})
        df['error'] = np.abs(df.preds - df.y)
        df_inrange = pd.DataFrame.copy(df.sort_values(by='error').iloc[:int(np.ceil(df.shape[0] * portion))])
        df_outrange = pd.DataFrame.copy(df.sort_values(by='error').iloc[int(np.ceil(df.shape[0] * portion)):])

        plt.scatter(df_inrange.y, df_inrange.preds, s=1, color="b", alpha=0.25)
        plt.scatter(df_outrange.y, df_outrange.preds, s=1, color="r", alpha=0.25)

        axes = plt.gca()
        y_min, y_max = axes.get_ylim()
        x_min, x_max = axes.get_xlim()
        start = min([y_min, x_min])
        end = min([y_max, x_max])

        line_values = np.arange(start, end, 0.01)
        plt.scatter(line_values,
                    line_values, alpha=1, s=2, color="w")

        plt.title("Predicted vs. observed (experimental) iRT - Highlight 95%")
        plt.xlabel("Observed iRT")
        plt.ylabel("Predicted iRT")
        plt.show()

        #plt.savefig(join(self._output_path, 'result_' + str(portion) + self._figures_ext))



import numpy as np

from dlomix.constants import retention_time_pipeline_parameters
from dlomix.data.RetentionTimeDataset import RetentionTimeDataset
from dlomix.models.base import RetentionTimePredictor


class RetentionTimePipeline:
    def __init__(self):
        super(RetentionTimePipeline, self).__init__()
        self.model = None
        self.dataset = None

        self._build_model()

    def _build_model(self):
        self.model = RetentionTimePredictor(**retention_time_pipeline_parameters['model_params'])
        self.model.load_weights(retention_time_pipeline_parameters['trained_model_path'])

    '''
    
    Predict retention times given data either as numpy array of sequences or a filepath to a csv file
    
    '''

    def predict(self, data=None):
        if not (isinstance(data, str) or isinstance(data, np.ndarray)):
            raise ValueError("Dataset should be provided either as a numpy array or a string pointing to a file.")

        self.dataset = RetentionTimeDataset(data, **retention_time_pipeline_parameters['data_params'], val_ratio=0,
                                            test=True)
        self.dataset.data_mean, self.dataset.data_std = retention_time_pipeline_parameters['trained_model_stats']

        predictions = self.model.predict(self.dataset.get_tf_dataset('test'))

        predictions = self.dataset.denormalize_targets(predictions)

        # here more reporting of results using the RTReport class

        return predictions



import pandas as pd
import numpy as np
import tensorflow as tf

'''
 TODO: check if it is better to abstract out a generic class for TF dataset wrapper, including:
 - splitting data logic
 - loading data logic
 '''


class RetentionTimeDataset:
    ATOM_TABLE = None
    SPLIT_NAMES = ["train", "val", "test"]
    BATCHES_TO_PREFETCH = tf.data.AUTOTUNE

    # TODO: For test dataset --> examples with longer sequences --> do not drop, add NaN for prediction

    def __init__(self, data_source=None, sep=",", sequence_col="sequence", target_col="irt", feature_cols=None,
                 normalize_targets=False, seq_length=0, batch_size=32, val_ratio=0, seed=21,
                 test=False, path_aminoacid_atomcounts=None, sample_run=False):
        super(RetentionTimeDataset, self).__init__()

        np.random.seed(seed)
        self.seed = seed

        self.data_source = data_source
        self.sep = sep
        self.sequence_col = sequence_col.lower()
        self.target_col = target_col.lower()
        if feature_cols:
            self.feature_cols = [f.lower() for f in feature_cols]
        else:
            self.feature_cols = []

        self.normalize_targets = normalize_targets
        self.sample_run = sample_run

        # if seq_length is 0 (default) -> no padding
        self.seq_length = seq_length

        self._data_mean = 0
        self._data_std = 1

        self.batch_size = batch_size
        self.val_ratio = val_ratio
        self.testing_mode = test

        self.main_split = RetentionTimeDataset.SPLIT_NAMES[2] if self.testing_mode else \
        RetentionTimeDataset.SPLIT_NAMES[0]

        self.sequences = None
        self.targets = None
        self.features_df = None

        # if path to counts lookup table is provided, include count features, otherwise not
        self.include_count_features = True if path_aminoacid_atomcounts else False

        if self.include_count_features:
            self.aminoacid_atom_counts_csv_path = path_aminoacid_atomcounts  # "../lookups/aa_comp_rel.csv"
            self._init_atom_table()

        # initialize TF Datasets dict
        self.tf_dataset = ({self.main_split: None, RetentionTimeDataset.SPLIT_NAMES[1]: None} if val_ratio != 0
                           else {self.main_split: None})

        self.indicies_dict = ({self.main_split: None, RetentionTimeDataset.SPLIT_NAMES[1]: None} if val_ratio != 0
                              else {self.main_split: None})

        # if data is provided with the constructor call --> load, otherwise --> done
        if self.data_source is not None:
            self.load_data(data=data_source)

    def _init_atom_table(self):
        atom_counts = pd.read_csv(self.aminoacid_atom_counts_csv_path)
        atom_counts = atom_counts.astype(str)

        keys_tensor = tf.constant(atom_counts['aa'].values)
        values_tensor = tf.constant(['_'.join(c) for c in list(atom_counts.iloc[:, 1:].values)])
        init = tf.lookup.KeyValueTensorInitializer(keys_tensor, values_tensor)
        RetentionTimeDataset.ATOM_TABLE = tf.lookup.StaticHashTable(init, default_value='0_0_0_0_0')

    def load_data(self, data):
        self.data_source = data

        self._read_data()
        self._validate_remove_long_sequences()
        self._split_data()
        self._build_tf_dataset()
        self._preprocess_tf_dataset()

    '''
    numpy array --> either a tuple or a single array
        - Tuple --> means (sequences, targets)
        - single ndarray --> means sequences only, useful for test dataset
    str --> path to csv file or compressed csv file
    '''

    def _read_data(self):
        if isinstance(self.data_source, tuple):
            tuple_size_is_two = len(self.data_source) == 2
            if tuple_size_is_two:
                tuple_elements_are_ndarray = (isinstance(self.data_source[0], np.ndarray)
                                              and isinstance(self.data_source[1], np.ndarray))
                if tuple_elements_are_ndarray:
                    self.sequences = self.data_source[0]
                    self.targets = self.data_source[1]
            else:
                raise ValueError(
                    'If a tuple is provided, it has to have a length of 2 and both elements should be numpy arrays.')

        elif isinstance(self.data_source, np.ndarray):
            self.sequences = self.data_source
            self.targets = np.zeros(self.sequences.shape[0])
            self._data_mean, self._data_std = 0, 1

        elif isinstance(self.data_source, str):
            df = pd.read_csv(self.data_source)
            if self.sample_run:
                df = df.head(100)

            # lower all column names
            df.columns = [col_name.lower() for col_name in df.columns]

            self.sequences, self.targets = df[self.sequence_col].values, df[self.target_col].values
            self._data_mean, self._data_std = np.mean(self.targets), np.std(self.targets)

            self.features_df = df[self.feature_cols]
        else:
            raise ValueError('Data source has to be either a tuple of two numpy arrays, a single numpy array, '
                             'or a string path to a csv file.')

    def _validate_remove_long_sequences(self) -> None:
        """
        Validate if all sequences are shorter than the padding length, otherwise drop them.
        """
        assert self.sequences.shape[0] > 0, "No sequences in the provided data."
        assert len(self.sequences) == len(self.targets), "Count of examples does not match for sequences and targets."

        limit = self.seq_length
        vectorized_len = np.vectorize(lambda x: len(x))
        mask = vectorized_len(self.sequences) <= limit
        self.sequences, self.targets = self.sequences[mask], self.targets[mask]

    def _split_data(self):
        n = len(self.targets)

        if self.val_ratio != 0 and (not self.testing_mode):
            self.indicies_dict[RetentionTimeDataset.SPLIT_NAMES[1]] = np.arange(n)[:int(n * self.val_ratio)]
            self.indicies_dict[self.main_split] = np.arange(n)[int(n * self.val_ratio):]
        else:
            self.indicies_dict[self.main_split] = np.arange(n)

    def _build_tf_dataset(self):
        for split in self.tf_dataset.keys():
            self.tf_dataset[split] = tf.data.Dataset.from_tensor_slices(
                (self.sequences[self.indicies_dict[split]], self.targets[self.indicies_dict[split]])
            )

    def _preprocess_tf_dataset(self):
        for split in self.tf_dataset.keys():
            # avoid normalizing targets for test data --> should not be needed
            if self.normalize_targets and not self.testing_mode:
                self.tf_dataset[split] = self.tf_dataset[split].map(lambda s, t: self._normalize_target(s, t),
                                                                    num_parallel_calls=tf.data.AUTOTUNE)

            self.tf_dataset[split] = (self.tf_dataset[split]
                                      .map(lambda s, t: self._split_sequence(s, t), num_parallel_calls=tf.data.AUTOTUNE)
                                      .map(lambda s, t: self._pad_sequences(s, t), num_parallel_calls=tf.data.AUTOTUNE)
                                      )

            if self.include_count_features:
                self.tf_dataset[split] = (self.tf_dataset[split]
                                          .map(RetentionTimeDataset._convert_inputs_to_dict,
                                               num_parallel_calls=tf.data.AUTOTUNE)
                                          .map(lambda s, t: self._generate_single_counts(s, t),
                                               num_parallel_calls=tf.data.AUTOTUNE)
                                          .map(lambda s, t: self._generate_di_counts(s, t),
                                               num_parallel_calls=tf.data.AUTOTUNE)
                                          )

            self.tf_dataset[split] = (self.tf_dataset[split]
                                      .batch(self.batch_size)
                                      .prefetch(RetentionTimeDataset.BATCHES_TO_PREFETCH)
                                      )

    def get_split_targets(self, split='val'):
        if split not in self.indicies_dict.keys():
            raise ValueError('requested split does not exist, availabe splits are: ' + list(self.indicies_dict.keys()))

        return self.targets[self.indicies_dict[split]]

    def denormalize_targets(self, targets):
        return targets * self._data_std + self._data_mean

    def _pad_sequences(self, seq, target):
        pad_len = tf.abs(self.seq_length - tf.size(seq))
        paddings = tf.concat([[0], [pad_len]], axis=0)
        seq = tf.pad(seq, [paddings], "CONSTANT")
        seq.set_shape([self.seq_length])
        return seq, target

    def _normalize_target(self, seq, target):

        target = tf.math.divide(tf.math.subtract(target, self._data_mean),
                                self._data_std)
        return seq, target

    def _split_sequence(self, seq, target):
        splitted_seq = tf.strings.bytes_split(seq)

        return splitted_seq, target

    '''
    if more than one input is added, inputs are added to a python dict, the following methods assume that
    '''

    @staticmethod
    def _convert_inputs_to_dict(seq, target):
        return {'seq': seq}, target

    def _generate_single_counts(self, inputs, target):
        inputs["counts"] = tf.map_fn(lambda x: RetentionTimeDataset.ATOM_TABLE.lookup(x), inputs["seq"])
        inputs["counts"] = tf.map_fn(lambda x: tf.strings.split(x, sep='_'), inputs["counts"])
        inputs["counts"] = tf.strings.to_number(inputs["counts"])
        inputs["counts"].set_shape([self.seq_length, 5])

        return inputs, target

    def _generate_di_counts(self, inputs, target):
        # add every two neighboring elements without overlap [0 0 1 1 2 2 .... pad_length/2 pad_length/2]
        segments_to_add = [i // 2 for i in range(self.seq_length)]
        inputs["di_counts"] = tf.math.segment_sum(inputs["counts"], tf.constant(segments_to_add))
        inputs["di_counts"].set_shape([self.seq_length // 2, 5])

        return inputs, target

    def get_tf_dataset(self, split=None):
        assert split in self.tf_dataset.keys(), (
            f"Requested data split is not available, available splits are {self.tf_dataset.keys()}")
        if split in self.tf_dataset.keys():
            return self.tf_dataset[split]
        return self.tf_dataset

    @property
    def train_data(self):
        return self.get_tf_dataset(RetentionTimeDataset.SPLIT_NAMES[0])

    @property
    def val_data(self):
        return self.get_tf_dataset(RetentionTimeDataset.SPLIT_NAMES[1])

    @property
    def test_data(self):
        return self.get_tf_dataset(RetentionTimeDataset.SPLIT_NAMES[2])

    @property
    def data_mean(self):
        return self._data_mean

    @property
    def data_std(self):
        return self._data_std

    @data_mean.setter
    def data_mean(self, value):
        self._data_mean = value

    @data_std.setter
    def data_std(self, value):
        self._data_std = value



import tensorflow as tf
import tensorflow.keras.backend as K

class TimeDeltaMetric(tf.keras.metrics.Metric):
    def __init__(self, mean=0, std=1, percentage=0.95, name='timedelta',
                 rescale_targets=False, rescale_predictions=False,
                 double_delta=False, **kwargs):
        super(TimeDeltaMetric, self).__init__(name=name, **kwargs)
        self.delta = self.add_weight(name='delta', initializer='zeros')
        self.batch_count = self.add_weight(name='batch-count', initializer='zeros')
        self.mean = mean
        self.std = std
        self.percentage = percentage
        self.rescale_targets = rescale_targets
        self.rescale_predictions = rescale_predictions
        self.double_delta = double_delta

    def update_state(self, y_true, y_pred, sample_weight=None):
        # rescale
        if self.rescale_targets:
            y_true = y_true * self.std + self.mean

        if self.rescale_predictions:
            y_pred = y_pred * self.std + self.mean

        # find position of the index
        length = tf.shape(y_true)[0]
        mark = tf.cast(length, dtype=tf.float32) * self.percentage
        mark = tf.cast(mark, dtype=tf.int32)

        # compute residuals and sort
        abs_error = tf.abs(y_true - y_pred)
        d = tf.sort(abs_error)[mark - 1]

        # two-sided delta
        if self.double_delta:
            d = d * 2

        # update count of batches
        self.batch_count.assign_add(1.0)

        # update delta
        self.delta.assign_add(tf.math.reduce_sum(d))

    def result(self):
        # this is simple averaging over the batches, more complex reduction can be added based on domain expertises
        # Examples are: take max or min of both deltas (translates to a strict or a relaxed metric)
        return tf.math.divide(self.delta, self.batch_count)


# code adopted and modified based on:
# https://github.com/horsepurve/DeepRTplus/blob/cde829ef4bd8b38a216d668cf79757c07133b34b/RTdata_emb.py
def delta95_metric(y_true, y_pred):
    mark95 = tf.cast(
        tf.cast(tf.shape(y_true)[0], dtype=tf.float32) * 0.95, dtype=tf.int32)
    abs_error = K.abs(y_true - y_pred)
    delta = tf.sort(abs_error)[mark95 - 1]
    norm_range = K.max(y_true) - K.min(y_true)
    return (delta * 2) / (norm_range)





import tensorflow as tf
from dlomix.constants import ALPHABET_UNMOD
from tensorflow.keras.layers.experimental import preprocessing


class DeepLCRetentionTimePredictor(tf.keras.Model):

    def __init__(self, seq_length=60, vocab_dict=ALPHABET_UNMOD, use_global_features=False):
        super(DeepLCRetentionTimePredictor, self).__init__()
        self.seq_length = seq_length
        self._use_global_features = use_global_features

        self.leaky_relu = tf.keras.layers.ReLU(max_value=20, negative_slope=0.1)
        self.string_lookup = preprocessing.StringLookup(vocabulary=list(vocab_dict.keys()))

        self._build_aminoacid_branch()
        self._build_diaminoacid_branch()
        self._build_onehot_encoding_branch()
        self._build_regressor()
        self.output_layer = tf.keras.layers.Dense(1)

        if self._use_global_features:
            self._build_global_features_branch()

    def _build_aminoacid_branch(self):
        self.aminoacid_branch = tf.keras.Sequential([
            self._build_conv_pool_block(n_filters=256, kernel=8, padding='same'),
            self._build_conv_pool_block(n_filters=128, kernel=8, padding='same'),
            self._build_conv_pool_block(n_filters=64, kernel=8, padding='same', pool=False),
            tf.keras.layers.Flatten()])

    def _build_diaminoacid_branch(self):

        self.diaminoacid_branch = tf.keras.Sequential([
            self._build_conv_pool_block(n_filters=128, kernel=2, padding='same'),
            self._build_conv_pool_block(n_filters=64, kernel=2, padding='same'),
            tf.keras.layers.Flatten()])

    def _build_global_features_branch(self):
        self.global_features_branch = tf.keras.Sequential([
            tf.keras.layers.Dense(16, activation=self.leaky_relu),
            tf.keras.layers.Dense(16, activation=self.leaky_relu),
            tf.keras.layers.Dense(16, activation=self.leaky_relu),

        ])

    def _build_onehot_encoding_branch(self):
        self.onehot_encoding_branch = tf.keras.Sequential([
            self._build_conv_pool_block(n_filters=2, kernel=2, padding='same', activation='tanh',
                                        pool_strides=10, pool_size=10),
            tf.keras.layers.Flatten()])

    def _build_regressor(self):
        self.regressor = tf.keras.Sequential([
            tf.keras.layers.Dense(128, activation=self.leaky_relu) for _ in range(5)])

    def _build_conv_pool_block(self, n_conv_layers=2, n_filters=256, kernel=8, padding="same", activation='leaky_relu',
                               pool=True, pool_strides=2, pool_size=2):

        # leaky relu by default
        activation_fn = self.leaky_relu

        if activation in ['tanh', 'relu']:
            activation_fn = activation

        block = tf.keras.Sequential([tf.keras.layers.Conv1D(filters=n_filters, kernel_size=kernel, padding=padding,
                                                            activation=activation_fn) for _ in range(n_conv_layers)])
        if pool:
            block.add(tf.keras.layers.MaxPooling1D(pool_size=2, strides=2))

        return block

    def call(self, inputs, **kwargs):
        outputs = {}

        integer_encoded = self.string_lookup(inputs['seq'])
        onehot_encoded = tf.one_hot(integer_encoded, depth=self.seq_length)

        if self._use_global_features:
            outputs['global_features_output'] = self.global_features_branch(inputs['global_features'])

        outputs['onehot_branch_output'] = self.onehot_encoding_branch(onehot_encoded)
        outputs['aminoacids_branch_output'] = self.aminoacid_branch(inputs['counts'])
        outputs['diaminoacids_branch_output'] = self.diaminoacid_branch(inputs['di_counts'])

        concatenated_output = tf.concat(outputs.values(), axis=1)
        concatenated_output = self.regressor(concatenated_output)
        return self.output_layer(concatenated_output)



import tensorflow as tf
from tensorflow.keras.layers.experimental import preprocessing
from dlomix.constants import ALPHABET_UNMOD
from dlomix.layers.attention import AttentionLayer


class PrositRetentionTimePredictor(tf.keras.Model):
    def __init__(self, embedding_output_dim=16, seq_length=30, vocab_dict=ALPHABET_UNMOD,
                 dropout_rate=0.5, latent_dropout_rate=0.1, recurrent_layers_sizes=(256, 512), regressor_layer_size=512):
        super(PrositRetentionTimePredictor, self).__init__()

        # tie the count of embeddings to the size of the vocabulary (count of amino acids)
        self.embeddings_count = len(vocab_dict) + 2

        self.dropout_rate = dropout_rate
        self.latent_dropout_rate = latent_dropout_rate
        self.regressor_layer_size = regressor_layer_size
        self.recurrent_layers_sizes = recurrent_layers_sizes

        self.string_lookup = preprocessing.StringLookup(vocabulary=list(vocab_dict.keys()))

        self.embedding = tf.keras.layers.Embedding(input_dim=self.embeddings_count,
                                                   output_dim=embedding_output_dim,
                                                   input_length=seq_length)
        self._build_encoder()

        self.attention = AttentionLayer()

        self.regressor = tf.keras.Sequential([
            tf.keras.layers.Dense(self.regressor_layer_size, activation='relu'),
            tf.keras.layers.Dropout(rate=self.latent_dropout_rate)
        ])

        self.output_layer = tf.keras.layers.Dense(1)

    def _build_encoder(self):
        self.encoder = tf.keras.Sequential([
            tf.keras.layers.Bidirectional(tf.keras.layers.GRU(units=self.recurrent_layers_sizes[0], return_sequences=True)),
            tf.keras.layers.Dropout(rate=self.dropout_rate),
            tf.keras.layers.GRU(units=self.recurrent_layers_sizes[1], return_sequences=True),
            tf.keras.layers.Dropout(rate=self.dropout_rate)
        ])

    def call(self, inputs, **kwargs):
        x = self.string_lookup(inputs)
        x = self.embedding(x)
        x = self.encoder(x)
        x = self.attention(x)
        x = self.regressor(x)
        x = self.output_layer(x)
        return x



import tensorflow as tf
from tensorflow.keras.layers.experimental import preprocessing
from dlomix.constants import ALPHABET_UNMOD


class RetentionTimePredictor(tf.keras.Model):

    def __init__(self, embedding_dim=16, seq_length=30,
                 encoder="conv1d", vocab_dict=ALPHABET_UNMOD):
        super(RetentionTimePredictor, self).__init__()

        # tie the count of embeddings to the size of the vocabulary (count of amino acids)
        self.embeddings_count = len(vocab_dict) + 2

        self.string_lookup = preprocessing.StringLookup(vocabulary=list(vocab_dict.keys()))

        self.embedding = tf.keras.layers.Embedding(input_dim=self.embeddings_count,
                                                   output_dim=embedding_dim,
                                                   input_length=seq_length)

        self._build_encoder(encoder)

        self.flatten = tf.keras.layers.Flatten()
        self.regressor = tf.keras.Sequential([
            tf.keras.layers.Dense(128, activation='relu'),
            tf.keras.layers.Dense(64, activation='relu')
        ])

        self.output_layer = tf.keras.layers.Dense(1)

    def _build_encoder(self, encoder_type):
        if encoder_type.lower() == 'conv1d':
            self.encoder = tf.keras.Sequential([
                tf.keras.layers.Conv1D(filters=256, kernel_size=3, padding='same', activation='relu'),
                tf.keras.layers.Conv1D(filters=512, kernel_size=3, padding='valid', activation='relu'),
                tf.keras.layers.MaxPooling1D(pool_size=2)])
        else:
            self.encoder = tf.keras.Sequential([
                tf.keras.layers.LSTM(256, return_sequences=True),
                tf.keras.layers.LSTM(256)
            ])

    def call(self, inputs, **kwargs):
        x = self.string_lookup(inputs)
        x = self.embedding(x)
        x = self.encoder(x)
        x = self.flatten(x)
        x = self.regressor(x)
        x = self.output_layer(x)

        return x




from dlomix.data import RetentionTimeDataset


def test_empty_rtdataset():
    rtdataset = RetentionTimeDataset()
    assert rtdataset.sequences is None
    assert rtdataset.targets is None
    assert rtdataset.main_split is RetentionTimeDataset.SPLIT_NAMES[0]



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='dlomix',
    version='0.0.1',
    author="Omar Shouman",
    author_email="o.shouman@tum.de",
    description="Deep Learning for Proteomics",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/wilhelm-lab/dlomix",
    packages=setuptools.find_packages(),
    install_requires=[
        'pandas',
        'numpy',
        'matplotlib',
        'scikit-learn',
        'tensorflow'],
    extras_require={
        "dev": [
            "pytest >= 3.7",
        ],
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Development Status :: 1 - Planning",
        "Intended Audience :: Science/Research"
    ],
)



from matplotlib import pyplot as plt
from os.path import join
from os import makedirs
import tensorflow as tf
import warnings
import abc

'''
    Base class for reports:
        - child classes should implement the abstract method generate_report 
'''


class Report(abc.ABC):
    VALID_FIGURE_FORMATS = ['pdf', 'jpeg', 'jpg', 'png']

    def __init__(self, output_path, history, figures_ext='pdf'):
        self._output_path = output_path
        makedirs(self._output_path, exist_ok=True)

        self._set_history_object(history)
        self._set_figures_format(figures_ext)

    def _set_history_object(self, history):
        # TODO: accept dictionary as well --> infer
        if not isinstance(history, tf.keras.callbacks.History):
            raise ValueError(
                'Reporting requires a History object (tf.keras.callbacks.History), which is returned from a call to '
                'model.fit(). Passed history argument is of type {} ', type(history)
            )
        if not hasattr(history, 'history'):
            raise ValueError(
                'The passed History object does not have a history attribute, which is a dict with results.'
            )

        if len(history.history.keys()) == 0:
            warnings.warn(
                'The passed History object contains an empty history dict, no training was done.'
            )

        self._history = history

    def _set_figures_format(self, figures_ext):
        figures_ext = figures_ext.lower()
        if figures_ext.startswith("."):
            figures_ext = figures_ext[1:]
        if figures_ext not in Report.VALID_FIGURE_FORMATS:
            raise ValueError("Allowed figure formats are: {}", Report.VALID_FIGURE_FORMATS)
        self._figures_ext = figures_ext

    '''
        Plot a keras metric given its name and the history object returned by model.fit() 
    '''

    def plot_keras_metric(self, metric_name):

        # add some checks on dict keys for the metrics, losses, etc...
        # otherwise throw an exception displaying available metrics

        plt.plot(self._history.history[metric_name])
        plt.plot(self._history.history['val_' + metric_name])
        plt.title(metric_name)
        plt.ylabel(metric_name)
        plt.xlabel('epoch')
        plt.legend(['train', 'val'], loc='upper left')
        plt.show()

        # TODO: check saving in terms of file size and format
        #plt.savefig(join(self._output_path, metric_name + '.' + self._figures_ext))

    def plot_all_metrics(self):
        metrics = self._history.history.keys()
        metrics = filter(lambda x: x.startswith("val_"), metrics)
        for metric in metrics:
            self.plot_keras_metric(metric)

    @abc.abstractmethod
    def generate_report(self, targets, predictions, **kwargs):
        pass


